import wntr
import collections
import numpy as np
from magnets.utils.call_on_functions import *

def parallel_pipes(relations, wn, new_link_list, junc_dict, pipe_dict, unremovable_nodes, special_nodes, special_links_nodes, special_links, alpha):
    connected_nodes = []
    num_connections = []
    num_junc = wn.num_junctions
    junc_names = wn.junction_name_list
    link_names = wn.link_name_list
    parallel_pipes_list = []
    
    for i in range(num_junc):
        connected_nodes.append([])
        for a,b in new_link_list:
            if (a == junc_names[i]):
                connected_nodes[i].append(b)
            if (b == junc_names[i]):
                connected_nodes[i].append(a)     
    
    for i in range(len(connected_nodes)):
        has_dup = ([item for item, count in collections.Counter(connected_nodes[i]).items() if count > 1])
        if len(has_dup)!= 0:
            # if junc_names[i] not in unremovable_nodes:
            if junc_names[i] not in special_nodes:
                for j in range(len(has_dup)):
                    # if has_dup[j] not in unremovable_nodes:
                    if has_dup[j] not in special_nodes:
                        if ((junc_names[i],has_dup[j]) not in parallel_pipes_list and (has_dup[j],junc_names[i]) not in parallel_pipes_list):
                            if ((junc_names[i],has_dup[j]) not in special_links_nodes and (has_dup[j],junc_names[i]) not in special_links_nodes):
                                parallel_pipes_list.append((junc_names[i],has_dup[j]))
                    
    parallel_links = []
    
    for j in range(len(parallel_pipes_list)):
        parallel_links.append([])
        a = parallel_pipes_list[j][0]
        b = parallel_pipes_list[j][1]
        
        for i in range(len(new_link_list)):
            if (new_link_list[i][0] == a and new_link_list[i][1] == b) or (new_link_list[i][1] == a and new_link_list[i][0] == b):
                if link_names[i] not in parallel_links and link_names[i] not in special_links:
                    parallel_links[j].append(link_names[i])
        
        # update junc_dict and relations to only reflect single pipes connecting two nodes
        junc_dict[a]['Connected nodes'] = list(np.unique(np.array((junc_dict[a]['Connected nodes']))))
        junc_dict[b]['Connected nodes'] = list(np.unique(np.array((junc_dict[b]['Connected nodes']))))
        relations[a] = list(np.unique(np.array((relations[a]))))
        relations[b] = list(np.unique(np.array((relations[b]))))
    
    # remove pipes in parallel and replace with single pipe
    for k in range(len(parallel_links)):
        leng = []
        ks = []
        
        for l in range(len(parallel_links[k])):
            pipe = wn.get_link(parallel_links[k][l])
            leng.append(pipe.length)
            ks.append(calc_K(pipe.length, pipe.diameter,pipe.roughness, alpha))
        
        new_l = min(leng)
        K_sum = 0
        
        for m in range(len(ks)):
            K_sum = K_sum + (1/ks[m])**(1/1.852)
            
        new_K = 1/(K_sum**1.852)
        new_d = (alpha*new_l/((100**1.852)*(new_K)))**(1/4.87)
        
        for n in range(len(parallel_links[k])):
            wn.remove_link(parallel_links[k][n], force=True)
            del pipe_dict[parallel_links[k][n]]
            
        wn.add_pipe('{}'.format(parallel_links[k][0]), start_node_name=parallel_pipes_list[k][0], end_node_name=parallel_pipes_list[k][1],length=new_l, diameter = new_d, roughness=100, minor_loss=0)
        pipe_dict[parallel_links[k][0]] = {'Start node name': parallel_pipes_list[k][0], 'End node name':parallel_pipes_list[k][1], 'Length': new_l, 'Diameter':new_d, 'Roughness':100}
        new_link_list.append((parallel_pipes_list[k][0],parallel_pipes_list[k][1]))
        
    return wn, junc_dict, pipe_dict, relations, new_link_list



import wntr
from magnets.utils.call_on_functions import *
from magnets.utils.characteristics import *

def reinitialize(wn, new_link_list, special_nodes, unremovable_nodes):
    junc_names = wn.junction_name_list
    
    link_1 = []
    link_2 = []
    
    for link_name, link in wn.links():
        link_1.append(link.start_node_name)
        link_2.append(link.end_node_name)
    link_list = tuple(zip(link_1,link_2))
    new_link_list = list(link_list)
    
    link_list_only_junc = []
    for a,b in new_link_list:
        if a not in special_nodes and b not in special_nodes:
            link_list_only_junc.append((a,b))
    
    for a,b in new_link_list:
        if (a in special_nodes and b not in special_nodes):
            if b not in unremovable_nodes:
                unremovable_nodes.append(b)
        if (b in special_nodes and a not in special_nodes):
            if a not in unremovable_nodes:
                unremovable_nodes.append(a)    
    
    nodes_to_be_removed = []
    for a in junc_names:
        if (a not in unremovable_nodes):
            nodes_to_be_removed.append(a)
    
    return nodes_to_be_removed, link_list_only_junc, unremovable_nodes



import wntr
from magnets.reduction import build_g_matrix
from magnets.reduction import reinitialize
from magnets.utils.characteristics import *
from magnets.utils.call_on_functions import *
import warnings

def mod_reduction(wn, new_link_list, junc_dict, pipe_dict, unremovable_nodes, relations, nodes_to_be_removed, new_pipe_len, alpha, inp_file, max_nodal_degree, op_pt):
    
    if max_nodal_degree == None: 
        max_nodal_degree = 100
        
    if (max_nodal_degree < 1 or (isinstance(max_nodal_degree, int) == False and max_nodal_degree != None)):
        warnings.warn('Invalid maximum nodal degree provided by user. \ Running simulation with maximum nodal degree = 100.')
        max_nodal_degree = 100
        
    count = 1
    
    while (len(nodes_to_be_removed)!=0):
               
        junc_names = wn.junction_name_list
        num_junc = len(junc_names)
        pipe_names = wn.pipe_name_list
        num_pipes = len(pipe_names)
                
        #identify node with minimum number of connections
        num_connections = []
        
        for i in range(len(nodes_to_be_removed)):
            num_connections.append(len(junc_dict[nodes_to_be_removed[i]]['Connected nodes']))
            
        removal_node = nodes_to_be_removed[num_connections.index(min(num_connections))]
        
        if min(num_connections) > max_nodal_degree:
            break
        
        neighbor_nodes = junc_dict[removal_node]['Connected nodes']
                
        if len(neighbor_nodes)==1:
            
            nb_node = neighbor_nodes[0]
            for j in range(num_pipes):
                if (pipe_dict[pipe_names[j]]['Start node name'] == nb_node and pipe_dict[pipe_names[j]]['End node name'] == removal_node) or (pipe_dict[pipe_names[j]]['Start node name'] == removal_node and pipe_dict[pipe_names[j]]['End node name'] == nb_node):
                    nb_link = pipe_names[j]
            
            #update diagonal g~
            junc_dict[nb_node]['Diagonal g'] = junc_dict[nb_node]['Diagonal g'] - junc_dict[removal_node]['Diagonal g']
            
            #update demand
            junc_dict[nb_node]['Demand'] = junc_dict[nb_node]['Demand'] + junc_dict[removal_node]['Demand']
            wn.get_node(nb_node).demand_timeseries_list[0].base_value = wn.get_node(nb_node).demand_timeseries_list[0].base_value + wn.get_node(removal_node).demand_timeseries_list[0].base_value
            
            #delete link and node and update dictionaries
            wn.remove_link(nb_link)
            wn.remove_node(removal_node)
            
            nodes_to_be_removed.remove(removal_node)
            del relations[removal_node]
            del junc_dict[removal_node]
            del pipe_dict[nb_link]
            
            junc_names = wn.junction_name_list
            num_junc = len(junc_names)
            
            for value in relations.values():
                if removal_node in value:
                    value.remove(removal_node)
            for k in range(num_junc):
                if removal_node in junc_dict[junc_names[k]]['Connected nodes']:
                    junc_dict[junc_names[k]]['Connected nodes'].remove(removal_node)
            
        else:
            
            removal_links = []
            pipe_names = wn.pipe_name_list
            
            for d in range(len(neighbor_nodes)):
                node_x = neighbor_nodes[d]
                for e in range(len(pipe_names)):
                    if (pipe_dict[pipe_names[e]]['Start node name'] == node_x and pipe_dict[pipe_names[e]]['End node name'] == removal_node) or (pipe_dict[pipe_names[e]]['Start node name'] == removal_node and pipe_dict[pipe_names[e]]['End node name'] == node_x):
                        link_x = pipe_names[e]
                        
                        #update demand
                        junc_dict[node_x]['Demand'] = junc_dict[node_x]['Demand'] + abs(pipe_dict[link_x]['Linear g']*junc_dict[removal_node]['Demand']/junc_dict[removal_node]['Diagonal g'])
                        wn.get_node(node_x).demand_timeseries_list[0].base_value = wn.get_node(node_x).demand_timeseries_list[0].base_value + abs(pipe_dict[link_x]['Linear g']*wn.get_node(removal_node).demand_timeseries_list[0].base_value/junc_dict[removal_node]['Diagonal g'])
 
                        #update diagonal g
                        junc_dict[node_x]['Diagonal g'] = junc_dict[node_x]['Diagonal g'] - abs((pipe_dict[link_x]['Linear g']**2)/junc_dict[removal_node]['Diagonal g'])
                        
                        removal_links.append(link_x)
            
            nb_pairs = [(neighbor_nodes[a],neighbor_nodes[b]) for a in range(len(neighbor_nodes)) for b in range(a+1, len(neighbor_nodes))]
            
            
            for node1, node2 in nb_pairs:
            
                link_1_2 = -1
                pipe_names = wn.pipe_name_list
                
                
                for l in range(len(pipe_names)):
                    
                    if (pipe_dict[pipe_names[l]]['Start node name'] == node1 and pipe_dict[pipe_names[l]]['End node name'] == node2) or (pipe_dict[pipe_names[l]]['Start node name'] == node2 and pipe_dict[pipe_names[l]]['End node name'] == node1):
                        link_1_2 = pipe_names[l]
                        
                    if (pipe_dict[pipe_names[l]]['Start node name'] == node1 and pipe_dict[pipe_names[l]]['End node name'] == removal_node) or (pipe_dict[pipe_names[l]]['Start node name'] == removal_node and pipe_dict[pipe_names[l]]['End node name'] == node1):
                        link_1_rem = pipe_names[l]
                        
                    if (pipe_dict[pipe_names[l]]['Start node name'] == node2 and pipe_dict[pipe_names[l]]['End node name'] == removal_node) or (pipe_dict[pipe_names[l]]['Start node name'] == removal_node and pipe_dict[pipe_names[l]]['End node name'] == node2):
                        link_2_rem = pipe_names[l]
                                                       
                #update non-diagonal g
                #Existing link between nodes 1 and 2
                if link_1_2 != -1:
                    pipe_dict[link_1_2]['Linear g'] = pipe_dict[link_1_2]['Linear g'] - abs(pipe_dict[link_1_rem]['Linear g']*pipe_dict[link_2_rem]['Linear g']/junc_dict[removal_node]['Diagonal g'])
                    pipe_dict[link_1_2]['Diameter'] = calculate_new_D(lin_g_to_nonlin_g(pipe_dict[link_1_2]['Linear g'],junc_dict[node1]['Head at op pt'], junc_dict[node2]['Head at op pt']),pipe_dict[link_1_2]['Length'],alpha,100)
                    pipe = wn.get_link(link_1_2)
                    pipe.diameter = pipe_dict[link_1_2]['Diameter']
                    pipe.roughness = 100
                    
                #No previously existing link between nodes 1 and 2
                else:
                    new_g = - abs(pipe_dict[link_1_rem]['Linear g']*pipe_dict[link_2_rem]['Linear g']/junc_dict[removal_node]['Diagonal g'])
                                        
                    wn.add_pipe('new-pipe-{}'.format(count), start_node_name = node1, end_node_name = node2, length = new_pipe_len, diameter = calculate_new_D(lin_g_to_nonlin_g(new_g,junc_dict[node1]['Head at op pt'], junc_dict[node2]['Head at op pt']), new_pipe_len, alpha, 100), roughness = 100, minor_loss = 0)
                    pipe_dict['new-pipe-{}'.format(count)] = {'Start node name': node1, 'End node name': node2, 'Length': new_pipe_len, 'Diameter':calculate_new_D(lin_g_to_nonlin_g(new_g,junc_dict[node1]['Head at op pt'], junc_dict[node2]['Head at op pt']), new_pipe_len, alpha, 100), 'Roughness':100, 'Linear g': new_g}
                    junc_dict[node1]['Connected nodes'].append(node2)
                    junc_dict[node2]['Connected nodes'].append(node1)
                    relations[node1].append(node2)
                    relations[node2].append(node1)
                    
                    count = count + 1

            for m in range(len(removal_links)):
                wn.remove_link(removal_links[m])
                del pipe_dict[removal_links[m]]                                  

            wn.remove_node(removal_node)
            
            nodes_to_be_removed.remove(removal_node)
            del relations[removal_node]
            del junc_dict[removal_node]
            
            junc_names = wn.junction_name_list
            num_junc = len(junc_names)
            
            for value in relations.values():
                if removal_node in value:
                    value.remove(removal_node)
            for k in range(num_junc):
                if removal_node in junc_dict[junc_names[k]]['Connected nodes']:
                    junc_dict[junc_names[k]]['Connected nodes'].remove(removal_node)
                    
    if '/' in inp_file:
        index = inp_file.rindex('/') + 1
        new_name = inp_file[:index] + 'reduced ' + str(op_pt) + ' ' + inp_file[index:] 
        wn.write_inpfile(new_name)
        
    else:
        wn.write_inpfile('reduced {} {}'.format(inp_file, op_pt))
                    
    return 1



import wntr
from wntr.network.elements import Pipe, Junction
import warnings


def characteristics(inp_file, op_pt = None, nodes_to_keep = None):
    
    wn = wntr.network.WaterNetworkModel(inp_file)
    sim = wntr.sim.EpanetSimulator(wn)
    results = sim.run_sim()
    
    report_time_step = wn.options.time.report_timestep
    
    horizon = int(wn.options.time.duration/report_time_step)
    time_steps = list(range(0,horizon+1))
    
    if op_pt == None:
        op_pt = 0
    
    if op_pt not in time_steps:
        warnings.warn('Invalid operating point provided by user. \ Running simulation with operating point = 0.')
        op_pt = 0
        
    #returns list of junction demands 
    dq = []
    
    for junc_name, junction in wn.junctions():
        dq.append(junction.base_demand)
        
    link_1 = []
    link_2 = []
    
    #build link_list and new_link_list
    for link_name, link in wn.links():
        n1 = link.start_node_name
        n2 = link.end_node_name
        link_1.append(link.start_node_name)
        link_2.append(link.end_node_name)
    link_list = tuple(zip(link_1,link_2))
    new_link_list = list(link_list)
    
    #building junction dictionary
    junc_names = wn.junction_name_list
    junc_dict = {}
    
    for i in range(len(junc_names)):
        junc_name = junc_names[i]
        junction = wn.get_node(junc_name)
        
        connected_nodes = []
        for a,b in new_link_list:
            if (a == junc_name):
                connected_nodes.append(b)
            if (b == junc_name):
                connected_nodes.append(a)

        op_pt_head = results.node['head'].loc[op_pt*report_time_step, junc_name]
            
        junc_dict[junc_name] = {'Demand': results.node['demand'].loc[op_pt*report_time_step, junc_name], 'Connected nodes': connected_nodes, 'Head at op pt': op_pt_head, 'Diagonal g': 0}
    
    #building pipe dictionary
    pipe_names = wn.pipe_name_list
    pipe_dict = {}
    
    for i in range(len(pipe_names)):
        pipe_name = pipe_names[i]
        pipe = wn.get_link(pipe_name)
        pipe_dict[pipe_name] = {'Start node name': pipe.start_node_name, 'End node name': pipe.end_node_name, 'Length': pipe.length, 'Diameter': pipe.diameter, 'Roughness':pipe.roughness, 'Linear g':0}
        
    # In[]: BUILD LISTS OF SPECIAL AND UNREMOVABLE ELEMENTS
        
    special_links = []
    special_nodes = []
    unremovable_nodes = []
    special_links_nodes = []
    
    #check reservoirs and tanks
    for res_name, reservoir in wn.reservoirs():
        if reservoir.name not in special_nodes:
            special_nodes.append(reservoir.name)
            unremovable_nodes.append(reservoir.name)
    
    for tank_name, tank in wn.tanks():
        if tank.name not in special_nodes:
            special_nodes.append(tank.name)
            unremovable_nodes.append(tank.name)
    
    #check nodes with negative demand
    for i in range(len(dq)):
        if dq[i]<0:
            if junc_names[i] not in special_nodes:
                special_nodes.append(junc_names[i])
                unremovable_nodes.append(junc_names[i])
    
    #check nodes connected to pumps and valves
    for pump_name, pump in wn.pumps():
        n1 = pump.start_node_name
        n2 = pump.end_node_name
        if n1 not in special_nodes:
            unremovable_nodes.append(n1)
        if n2 not in special_nodes:
            unremovable_nodes.append(n2)
        special_links.append(pump_name)
        special_links_nodes.append((n1,n2))
        
    for valve_name, valve in wn.valves():
        n1 = valve.start_node_name
        n2 = valve.end_node_name
        if n1 not in special_nodes:
            unremovable_nodes.append(n1)
        if n2 not in special_nodes:
            unremovable_nodes.append(n2)
        special_links.append(valve_name)
        special_links_nodes.append((n1,n2))
    
    #identify junctions and links with control rules
    junc_with_controls = []
    pipe_with_controls = []
    for name, control in wn.controls():
            for req in control.requires():
                if isinstance(req, Junction):
                    junc_with_controls.append(req.name)
                elif isinstance(req, Pipe):
                    pipe_with_controls.append(req.name)
    junc_with_controls = list(set(junc_with_controls))
    pipe_with_controls = list(set(pipe_with_controls))
    
    #set control junctions and links to be unremovable
    for junc_name in junc_with_controls:
        if junc_name not in unremovable_nodes:
            unremovable_nodes.append(junc_name)
            
    for pipe_name in pipe_with_controls:
        n1 = wn.get_link(pipe_name).start_node_name
        n2 = wn.get_link(pipe_name).end_node_name
        if n1 not in special_nodes and n1 not in unremovable_nodes:
            unremovable_nodes.append(n1)
        if n2 not in special_nodes and n2 not in unremovable_nodes:
            unremovable_nodes.append(n2)
        special_links.append(pipe_name)
        special_links_nodes.append((n1,n2))
    
    #check nodes connected to special nodes          
    for a,b in new_link_list:
        if (a in special_nodes and b not in special_nodes):
            if b not in unremovable_nodes:
                unremovable_nodes.append(b)
        if (b in special_nodes and a not in special_nodes):
            if a not in unremovable_nodes:
                unremovable_nodes.append(a)    
                
    #check list of nodes to keep provided by user
    if nodes_to_keep is not None:
        counter = 0
        for i in range(len(nodes_to_keep)):
            if nodes_to_keep[i] not in unremovable_nodes and nodes_to_keep[i] not in special_nodes and nodes_to_keep[i] in junc_names:
                unremovable_nodes.append(nodes_to_keep[i])
            else:
                counter = counter+1
        if counter > 0:
            warnings.warn('Some values in list of nodes to keep provided by user do not exist in the model or have already been classified as special nodes. These values have been ignored.')
                    
    #build relations
    connected_nodes = []
    num_connections = []
    
    for i in range(len(junc_names)):
        connected_nodes.append([])
        for a,b in new_link_list:
            if (a == junc_names[i]):
                connected_nodes[i].append(b)
            if (b == junc_names[i]):
                connected_nodes[i].append(a)
    
    relations = dict(zip(junc_names,connected_nodes))
    
    return wn, op_pt, pipe_dict,junc_dict,unremovable_nodes,special_nodes,special_links,special_links_nodes, relations, new_link_list, results



import numpy as np

epsilon = 10**(-6)

#function to calculate alpha
def calculate_unit_coeff(pipe_dict, junc_dict, wn, results, op_pt, special_nodes):
    alpha = []
    pipe_names = wn.pipe_name_list
    
    report_time_step = wn.options.time.report_timestep
    
    for i in range(len(wn.pipe_name_list)):
        if (results.link['flowrate'].loc[op_pt*report_time_step, pipe_names[i]]>0):
            flow = results.link['flowrate'].loc[op_pt*report_time_step, pipe_names[i]]
            pipe = pipe_names[i]
            n1 = pipe_dict[pipe]['Start node name']
            n2 = pipe_dict[pipe]['End node name']
            if n1 not in special_nodes and n2 not in special_nodes:
                h1 = junc_dict[n1]['Head at op pt']
                h2 = junc_dict[n2]['Head at op pt']
                alpha.append(((abs(h1-h2))*((pipe_dict[pipe]['Roughness']/flow)**1.852)*(pipe_dict[pipe]['Diameter']**4.87))/pipe_dict[pipe]['Length'])
    
    return (np.mean(np.array(alpha)))

# alpha = calculate_unit_coeff(pipe_dict, junc_dict, wn)

#function to calulate K
def calc_K(L,D,R,alpha):
    return (alpha*L)/((R**1.852)*(D**4.87))

#function to calculate non-linear g
def nonlin_g(leng,diam,rough,alpha):
    K = 0
    g = 0
    K = calc_K(leng,diam,rough,alpha)
    g = (1/K)**(1/1.852)
    return g

#function to convert non-linear to linearized g
def nonlin_g_to_lin_g(g, h1, h2):
    if h1 == h2:
        h1 = 0
        h2 = epsilon
    return ((g*(abs(h1-h2)**((1/1.852)-1)))/1.852)

#function to convert linearized to non-linear g
def lin_g_to_nonlin_g(g,h1,h2):
    if h1 == h2:
        h1 = 0
        h2 = epsilon
    return ((abs(g)*1.852)/(abs(h1-h2)**((1/1.852)-1)))

#function calculate diameter of each new pipe
def calculate_new_D(g, L, alpha, C):
    return ((((g/C)**1.852)*alpha*L)**(1/4.87))

#function to calculate pipe length of all new pipes
def new_pipe_length(wn, pipe_dict):
    lengths = []
    pipe_names = wn.pipe_name_list
    for i in range(len(wn.pipe_name_list)):
        lengths.append(pipe_dict[pipe_names[i]]['Length'])
    return (np.mean(np.array(lengths)))
